#include "swplatform.h"

#include <timers.h>

#define C_SCAN_ROWS 8
#define C_SCAN_COLS 8
#define C_RET_MASK (1<<C_SCAN_COLS)-1
#define C_SCAN_MASK (1<<C_SCAN_ROWS)-1
#define C_DISP_PERIOD 2

#define C_ACT_CLOSE 129
#define C_ACT_OPEN  130
#define C_ACT_DUMP  299

typedef struct ent {
    struct ent * nextptr;
    struct ent * prevptr;
} ent;

typedef struct keymsg {
    struct ent * nextptr;
    struct ent * prevptr;
    int row;
    int column;
    int action;
} keymsg;

struct ent key_queue;

int ub_state = 0;
int timer_ticks = 0;
int alive_counter = 0;
int display_period = 0;
int alive_led_status = 0;
int scanrow = 0;
int scanstat1[C_SCAN_ROWS];
int scanstat2[C_SCAN_ROWS];

// error starting tick timer
#define C_ERR_TREG 130
// key message buffer space exhausted
#define C_ERR_KMEX 131

//
// timer shall tick in 10msec intervals
//
#define C_TICK_LENGTH 10000
//
// show alive sign every 1000 msec
//
#define C_ALIVE_INTERVAL 100

void error_halt(int errcod) {
    led_turn_all_off(drv_led_1);
    int i;
    for (i=0; i<=7; i++) {
      if (errcod&1)
        led_turn_on(drv_led_1,LEDS_LED0_R+(i*3));
      else
        led_turn_off(drv_led_1,LEDS_LED0_R+(i*3));
      errcod /= 2;
    }
    while (1);
}
void init_leds() {
    led_turn_all_off(drv_led_1);
    led_turn_on(drv_led_1,LEDS_LED0_B);
}

void init_queue(struct ent * queptr) {
    queptr->nextptr=0;
    queptr->prevptr=0;
}

int queue_has_entries(struct ent * queptr) {
    if (queptr->nextptr!=0) {
        return true;
    }
    else {
        return false;
    }
}

void insert_queue_last(struct ent * queptr, struct ent * entptr) {
    if (queptr->nextptr!=0) {
        entptr->nextptr=queptr->prevptr;
        queptr->prevptr->prevptr=entptr;
        queptr->prevptr=entptr;
        entptr->prevptr=0;
    }
    else {
        queptr->nextptr=entptr;
        queptr->prevptr=entptr;
        entptr->nextptr=0;
        entptr->prevptr=0;
    }
}

ent * get_queue_first(struct ent * queptr) {
    if (!queue_has_entries(queptr)) {
        return 0;
    }
    else {
        struct ent * entptr;
        entptr = queptr->nextptr;
        queptr->nextptr=queptr->nextptr->nextptr;
        return entptr;
    }
}

void check_alive() {
    alive_counter++;
    display_period++;
    if (alive_counter>=C_ALIVE_INTERVAL) {
        alive_counter = 0;
        if (alive_led_status==0) {
            alive_led_status = 1;
            led_turn_on(drv_led_1, LEDS_LED7_G);
        }
        else {
            alive_led_status = 0;
            led_turn_off(drv_led_1, LEDS_LED7_G);
        }
        if (display_period>=C_DISP_PERIOD) {
            struct keymsg * kmptr = malloc(sizeof(struct keymsg));
            kmptr->action=C_ACT_DUMP;
            kmptr->row=0;
            kmptr->column=0;
            insert_queue_last(&key_queue,(struct ent*)kmptr);
            display_period=0;
        }
    }
}

void init_scanpoints() {
    scanrow = 0;
    int i = 0;
    for (i=0;i<C_SCAN_ROWS;i++) {
        scanstat1[i]=0;
        scanstat2[i]=0;
        ioport_set_value(drv_ioport_1,KEY_MATRIX_SCAN,1<<scanrow);
    }
}

void check_scanpoints() {
    int krow = ioport_get_value(drv_ioport_1,KEY_MATRIX_RET);
    krow ^= 0x0FF;
    krow &= C_RET_MASK;
    krow &= scanstat1[scanrow];
    if (scanstat1[scanrow]!=scanstat2[scanrow]) {
      int scancol=0;
      int keystat=0;
      int oldstat=0;
      int newstat=0;
      int colmask=0;
      for (scancol=0; scancol<C_SCAN_COLS; scancol++) {
        colmask = 1<<scancol;
        oldstat = scanstat2[scanrow] & colmask;
        newstat = scanstat1[scanrow] & colmask;
        if (oldstat!=newstat) {
          struct keymsg * kmptr = malloc(sizeof(struct keymsg));
          if (!kmptr) {
            error_halt(C_ERR_KMEX);
          }
          kmptr->row=scanrow;
          kmptr->column=scancol;
          if (newstat!=0) {
            kmptr->action = C_ACT_CLOSE;
          }
          else {
            kmptr->action = C_ACT_OPEN;
          }
          insert_queue_last(&key_queue,(struct ent*)kmptr);
        }
      }
      scanstat2[scanrow]=scanstat1[scanrow];
    }
    scanrow++;
    if (scanrow>=C_SCAN_ROWS) {
        scanrow=0;
    }
    ioport_set_value(drv_ioport_1,KEY_MATRIX_SCAN,((1<<scanrow)^C_SCAN_MASK));
}

void tick(void * timerctx) {
    timer_ticks++;
    // timer_register_handler(0,C_TICK_LENGTH,tick);
    check_alive();
    check_scanpoints();
}

void init_ub() {
    ub_state = 0;
    timer_ticks=0;
    void * timer_id = timer_register_handler((void*)0,C_TICK_LENGTH,tick);
    if (!timer_id) {
        error_halt(C_ERR_TREG);
    }
    init_scanpoints();
    init_queue(&key_queue);
}

void check_ub() {
    int ub_act = ioport_get_value(drv_ioport_2,0);
    int i;
    for (i=0; i<5; i++) {
        if ((ub_act & (1<<i)) != 0)
          led_turn_on(drv_led_1,LEDS_LED1_B+(3*i));
        else
          led_turn_off(drv_led_1,LEDS_LED1_B+(3*i));
    }
}

void check_event() {
    if (queue_has_entries(&key_queue)) {
      keymsg * kmp = (keymsg *) get_queue_first(&key_queue);
      if (kmp) {
        printf("Key Event Message: row=%d, col=%d, action=%d\n",kmp->row,kmp->column,kmp->action);
        switch (kmp->action) {
        case C_ACT_DUMP:
          printf("Key Matrix Status:\n");
          int row=0;
          int col=0;
          for(row=0;row<C_SCAN_ROWS;row++) {
            printf("%d - ",row);
            for(col=0;col<C_SCAN_COLS;col++) {
              if (((1<<col)&scanstat1[row])!=0) {
                printf("*");
              }
              else {
                printf("_");
              }
            }
            printf("\n");
          }
          break; // case C_ACT_DUMP
        } // switch
      } // if kmp
    } // if has_entries
} // check_event

void main() {
  swplatform_init_stacks();
  init_leds();
  init_ub();
  led_turn_off(drv_led_1,LEDS_LED0_B);
  led_turn_on(drv_led_1,LEDS_LED0_G);
  printf("MIDIKBD initialized\n");
  while (1) {
    check_ub();
    check_event();
  }
}
